import logging
from pathlib import Path

import natsort as ns

from scisort.api import Group
from scisort.api import Pattern

SCISORT_DEFAULT = Group(
    [
        # level: Landing docs
        Group([Pattern(regexp=r"read[-_\s]{0,1}me.*")], name="doc_landing"),
        # level: License
        Group([Pattern(regexp=r"license.*")], name="license"),
        # level: Installation
        Group(
            [
                Pattern(regexp=r"installation.*"),
                Pattern(regexp=r"install.*"),
                Pattern(regexp=r"setup.*"),
                Pattern(s="requirements.txt"),
            ],
            name="installation",
        ),
        # level: Citation
        Group([Pattern(regexp=r"citation.*")], name="citation"),
        # level: Config & hidden
        Pattern(regexp=r"\..*"),
        # level: All other files without extensions
        Pattern(regexp=r".*\..*", stop_search=False),
        # level: Executables
        Group(
            [
                Pattern(regexp=r".*\.sh"),
                Pattern(regexp=r".*\.bat"),
                Pattern(regexp=r".*\.bash"),
                Pattern(regexp=r".*\.exe"),
            ],
            name="executable",
        ),
        # level: Data
        Group(
            [
                Group(
                    [
                        Pattern(regexp=r"raw[-_\s]{0,1}data.*"),
                        Pattern(regexp=r"data[-_\s]{0,1}raw.*"),
                    ]
                ),
                Group(
                    [
                        Pattern(regexp=r"clean[-_\s]{0,1}data.*"),
                        Pattern(regexp=r"data[-_\s]{0,1}clean.*"),
                    ]
                ),
                Pattern(regexp=r"data.*"),
            ],
            name="data",
        ),
        # level: Scripts
        Group(
            [
                Pattern(regexp=r"scripts.*"),
                Pattern(regexp=r"src.*"),
                Pattern(regexp=r"code.*"),
            ],
            name="scripts",
        ),
        # level: Results
        Group(
            [Pattern(regexp=r"output.*"), Pattern(regexp=r"results.*")], name="results"
        ),
        # level: Tests
        Group([Pattern(regexp=r"test.*")], name="tests"),
    ]
)


def scisort_keygen(pattern=SCISORT_DEFAULT, **kwargs):
    """Key for scientific file sorting.

    Arguments
    ---------
    pattern: scisort.api.Group
        Sort pattern or algorithm composed from groups and
        Patterns.
    **kwargs:
        Additional keyword arguments to be passed to
        natsort.natsort_keygen.

    Returns
    -------

    type:
        A function that parses input for scientific sorting that
        is suitable for passing as the key argument to functions
        such as sorted.
    """

    def _scisort_key(f, pattern=SCISORT_DEFAULT, **kwargs):
        def _matcher(f_sub, gr, rank=tuple()):

            if not isinstance(gr, Group):
                raise ValueError("Expected Group")

            latest_k = None

            # iter over patterns
            for rank_group, match_obj in enumerate(gr.match_objs):

                if isinstance(match_obj, Group):
                    k = _matcher(
                        f_sub,
                        match_obj,
                        rank=rank + (rank_group,),
                    )

                    if k and match_obj.stop_search:
                        latest_k = k
                        return latest_k

                # the object is a pattern
                if isinstance(match_obj, Pattern):

                    # there is a match with the file/folder
                    if match_obj.score(f_sub):

                        # set the result of the sorting to k
                        latest_k = (
                            (
                                rank + (rank_group,),
                                ns.natsort_keygen(alg=ns.PATH, **kwargs)(f_sub),
                            ),
                        )

                        # stop the search if given
                        if match_obj.stop_search:
                            return latest_k

            return latest_k

        res = tuple()
        for fpart in Path(f).parts:

            m = _matcher(fpart, pattern)

            if m is None:
                k = (
                    (
                        (len(pattern.match_objs),),
                        ns.natsort_keygen(alg=ns.PATH, **kwargs)(fpart),
                    ),
                )
                res = res + k
                continue

            res = res + m

        logging.debug(res)

        return res

    return _scisort_key


def scisort_keygen_pandas(**kwargs):
    """Key for scientific file sorting in Pandas.

    Arguments
    ---------
    pattern: scisort.api.Group
        Sort pattern or algorithm composed from groups and
        Patterns.
    **kwargs:
        Additional keyword arguments to be passed to
        natsort.natsort_keygen.

    Returns
    -------

    type:
        A function that parses input for scientific sorting that
        is suitable for passing as the key argument to pandas sort
        functions.
    """

    def _scisort_key_pandas(s, **kwargs):

        return s.map(scisort_keygen(**kwargs))

    return _scisort_key_pandas



import pandas as pd
from pandas.testing import assert_series_equal

from scisort import scisort_keygen
from scisort import scisort_keygen_pandas

tree = [
    ".flake8",
    "data",
    "data/a (11).csv",
    "data/a (8).csv.gz",
    "data/readme.md",
    "jobs.sh",
    "random_file.xyz",
    "README.md",
    "scripts",
    "installation.R",
    "requirements.txt",
    "tests",
]

tree_expected = [
    "README.md",
    "installation.R",
    "requirements.txt",
    ".flake8",
    "random_file.xyz",
    "jobs.sh",
    "data",
    "data/readme.md",
    "data/a (8).csv.gz",
    "data/a (11).csv",
    "scripts",
    "tests",
]


def test_keygen_readme():

    assert scisort_keygen()("readme.md")[0][0][0] == 0


def test_keygen_jobs():

    assert scisort_keygen()("jobs.sh")[0][0][0] == 6


def test_keygen_no_match():

    assert scisort_keygen()("jobs.xyz")[0][0][0] == 5


def test_keygen_data():

    assert scisort_keygen()("data")[0][0][0] == 7


def test_keygen_sorted():

    assert sorted(tree, key=scisort_keygen()) == tree_expected


def test_keygen_pandas():

    s = pd.Series(tree)
    s_exp = pd.Series(tree_expected)

    assert_series_equal(
        s.sort_values(ignore_index=True, key=scisort_keygen_pandas()), s_exp
    )



import re


class Pattern(object):
    """File or folder object to match on."""

    def __init__(self, s=None, regexp=None, lower=True, stop_search=True):
        super(Pattern, self).__init__()
        self.s = s
        self.regexp = regexp
        self.lower = lower
        self.stop_search = stop_search

    def score(self, fp):

        if self.regexp:

            if self.lower:
                flags = re.IGNORECASE

            p = re.compile(self.regexp, flags=flags)

            return p.match(fp)

        if self.s:

            return isinstance(fp, str) and fp == self.s

        raise ValueError("Can't compute score")


class Group(object):
    """docstring for Group"""

    def __init__(self, match_objs, name=None, stop_search=True):
        super(Group, self).__init__()
        self.match_objs = match_objs
        self.name = name
        self.stop_search = stop_search



try:
    from scisort._version import __version__
    from scisort._version import __version_tuple__
except ImportError:
    __version__ = "0.0.0"
    __version_tuple__ = (0, 0, 0)


from scisort.keygen import scisort_keygen
from scisort.keygen import scisort_keygen_pandas

__all__ = ["scisort_keygen", "scisort_keygen_pandas"]
